"""Utilities for cleaning data."""
import os

import numpy as np
import pandas as pd
from functools import partial

def clean_fio2(df):
    # unphysiological data
    idx = df['FiO2'] <= 0.2
    df.loc[idx, 'FiO2'] = np.nan

    # proportion instead of percentage
    idx = df['FiO2'] <= 1.0
    df.loc[idx, 'FiO2'] = df.loc[idx, 'FiO2'] * 100.0

    # o2 flow inserted into FiO2
    idx = (df['FiO2'] > 1) & (df['FiO2'] <= 20)
    df.loc[idx, 'FiO2'] = np.nan

    # unphysiological data
    idx = df['FiO2'] >= 100.0
    df.loc[idx, 'FiO2'] = np.nan

    return df


def reject_outlier(low, high, value):
    """Reject anything outside the given boundaries."""
    if value < low:
        return None
    elif value > high:
        return None
    else:
        return value

def clean_instructions():
    """Hard-coded mapping of data type to cleaning function."""
    return {
        'Albumin': partial(reject_outlier, low=1, high=10),
        'Base Excess': partial(reject_outlier, low=0, high=30),
        'Bilirubin Serum': partial(reject_outlier, low=1, high=30),
        "Calcium, Serum": partial(reject_outlier, low=1, high=20),
        "Glucose, Serum Quant": partial(reject_outlier, low=0, high=1000),
        'PaO2': partial(reject_outlier, low=1, high=400),
        "PO2": partial(reject_outlier, low=1, high=400),
        'SaO2': partial(reject_outlier, low=5, high=100),
        "SaO2 %, Arterial": partial(reject_outlier, low=5, high=100),
        "SO2": partial(reject_outlier, low=5, high=100),
        'SPO2': partial(reject_outlier, low=5, high=100),
        'Pulse Oximetry': partial(reject_outlier, low=5, high=100),
        'Resp Rt Tot': partial(reject_outlier, low=5, high=70),
        'Resp Rt': partial(reject_outlier, low=5, high=70),
        'DBP': partial(reject_outlier, low=5, high=300),
        'MAP': partial(reject_outlier, low=5, high=200),
        'SBP': partial(reject_outlier, low=5, high=300),
        # heart rate
        'HR': partial(reject_outlier, low=1, high=300),
        'HR Monitored': partial(reject_outlier, low=1, high=300),
        'Pulse': partial(reject_outlier, low=1, high=300),
        'Pulse Peripheral': partial(reject_outlier, low=1, high=300),
        'HR Apical': partial(reject_outlier, low=1, high=300),
        'Temp': partial(reject_outlier, low=5, high=40),
        'Temp Axillary': partial(reject_outlier, low=5, high=40),
        'Temp Oral': partial(reject_outlier, low=5, high=40),
        'Temp Tympanic': partial(reject_outlier, low=5, high=40),
        'Temp Temporal Artery': partial(reject_outlier, low=5, high=40),
        'Height': partial(reject_outlier, low=0, high=220),
        'Weight': partial(reject_outlier, low=0, high=350),
        'Pain Score': partial(reject_outlier, low=0, high=10),
        "FLACC Pain Scale": partial(reject_outlier, low=0, high=10),
        # hematology cell %s
        'Lymph %': partial(reject_outlier, low=0, high=100),
        'Mono %': partial(reject_outlier, low=0, high=100),
        'Baso %': partial(reject_outlier, low=0, high=100),
        'Neutrophil %': partial(reject_outlier, low=0, high=100),
        'Diff, Eosinophil %': partial(reject_outlier, low=0, high=100),
        'Diff, Granulocyte %': partial(reject_outlier, low=0, high=100),
        'Diff Bands': partial(reject_outlier, low=0, high=100),
        'Imm Granulocytes Cnt Bld Auto, %': partial(reject_outlier, low=0, high=100),
        # hematology cell counts
        'Lymph Abs Cnt': partial(reject_outlier, low=0, high=1000)
    }

def clean_features(df):
    n = 0

    clean_fcns = clean_instructions()
    for col, fcn in clean_fcns.items():
        # skip if not present in df
        if col not in df.columns:
            continue
        
        # apply function that returns None if value is invalid
        df[r] = df[r].apply(fcn)

    # custom outlier rules that use more than just the data value itself

    # weights: (1) remove values with a min 50kg lower than the max
    wt_ids = set(df.loc[df['Weight'] > 270].index)
    wt_min_max = df.loc[wt_ids, 'Weight'].reset_index().groupby(
        df.index.name)['Weight'].agg(['min', 'max', 'count'])
    wt_ids = wt_min_max.loc[(wt_min_max['min'] + 50) < wt_min_max['max']]
    for adm_id in wt_ids.index:
        wt_rem = wt_ids.loc[adm_id, 'min'] + 50
        df.loc[adm_id, 'Weight'] = df.loc[adm_id, 'Weight'].apply(
            lambda x: x if x <= wt_rem else np.nan)

    # remaining patients have their weight converted to kg
    wt_ids = set(df.loc[df['Weight'] > 270].index)
    df.loc[wt_ids, 'Weight'] = df.loc[wt_ids,
                                      'Weight'].apply(lambda x: 0.453592*x)

    return df



# Generate the tables from BQ through python
import os
import sys
from importlib.resources import open_text
from google.cloud import storage, bigquery
from google.cloud.exceptions import NotFound, Forbidden, BadRequest
import logging
from pathlib import Path

import clinical_benchmarks

logger = logging.getLogger(__name__)


def check_bq_components(project, dataset, bucket_name):
    """ Check BQ components exists and error if they do not"""
    client_bq = bigquery.Client(project=project)
    try:
        client_bq.get_dataset(f'{project}.{dataset}')
    except BadRequest:
        logger.error(f'Project {project} not found')
    except NotFound:
        logger.error(f'Dataset {dataset} is not found in project {project}')

    client_storarge = storage.Client(project=project)
    bucket = client_storarge.bucket(bucket_name)
    try:
        bucket.exists()
    except BadRequest:
        logger.error(f'Bucket {bucket_name} does not exist')
    except Forbidden:
        logger.error(f'You do not have access to Bucket {bucket_name}')


def run_queries(project, dataset):
    """Run queries and output to a specified BigQuery dataset."""
    # Construct a BigQuery client object.
    client = bigquery.Client(project=project)

    # all pre-written SQL query in the qeury folder
    query_locations = [
        ['clinical_benchmarks.query.mimic', 'cohort.sql'],
        ['clinical_benchmarks.query.mimic.task_cohort', 'vancomycin_cohort.sql'],
        ['clinical_benchmarks.query.mimic.task_cohort', 'heparin_cohort.sql'],
        ['clinical_benchmarks.query.mimic.feature', 'chemistry.sql'],
        ['clinical_benchmarks.query.mimic.feature', 'coagulation.sql'],
        ['clinical_benchmarks.query.mimic.feature', 'complete_blood_count.sql'],
        ['clinical_benchmarks.query.mimic.feature', 'height.sql'],
        ['clinical_benchmarks.query.mimic.feature', 'weight.sql'],
        ['clinical_benchmarks.query.mimic.feature', 'vitalsign.sql'],
        ['clinical_benchmarks.query.mimic.outcome', 'vancomycin_dose.sql'],
        ['clinical_benchmarks.query.mimic.outcome', 'vancomycin_level.sql'],
        ['clinical_benchmarks.query.mimic.outcome', 'labevents.sql'],
        ['clinical_benchmarks.query.mimic.outcome', 'mortality.sql'],
        ['clinical_benchmarks.query.mimic.outcome', 'los_icu.sql'],
        ['clinical_benchmarks.query.mimic.outcome', 'dialysis.sql'],
        ['clinical_benchmarks.query.mimic.outcome', 'inv_vent.sql'],
        ['clinical_benchmarks.query.mimic.outcome', 'heparin_dose.sql']
    ]

    for module, query_name in query_locations:
        with open_text(module, query_name) as fp:
            query = ''.join(fp.readlines())

        query_stem = query_name.split('.')[0]
        table_name = f'{project}.{dataset}.{query_stem}'

        logger.info('Generating tables: ' + query_stem)
        # Overwrite table if exists (may remove this)
        job_config = bigquery.QueryJobConfig(
            destination=table_name, write_disposition='WRITE_TRUNCATE'
        )

        # call API to create the table
        query_job = client.query(query, job_config=job_config)
        # wait for the job to complete
        query_job.result()


def copy_from_bq_to_bucket(filename, project, dataset, bucket):
    """Copy a table from BigQuery to a bucket."""
    client = bigquery.Client(project=project)

    destination_uri = f'gs://{bucket}/{filename}'
    dataset = bigquery.DatasetReference(project, dataset)
    table = dataset.table(filename.split('.')[0])

    # extract gzip compressed CSV files
    config = bigquery.job.ExtractJobConfig(
        compression='gzip', destination_format='csv'
    )
    extract_job = client.extract_table(
        table, destination_uri, location="US", job_config=config
    )

    # wait for job to complete
    extract_job.result()
    logger.info(f'Created {destination_uri}')


def download_gcs_file(filename, csv_dir:Path, project, bucket_name='data-benchmark'):
    logger.info(f'Downloading: {filename}')
    file_path = csv_dir / filename
    client = storage.Client(project=project)
    bucket = client.bucket(bucket_name)

    blob = bucket.blob(filename)
    if not os.path.exists(csv_dir):
        os.makedirs(csv_dir)

    blob.download_to_filename(file_path)



import pandas as pd
from pathlib import Path
import logging

from clinical_benchmarks.merge import join_cohort_data

logger = logging.getLogger(__name__)

# read the downloaded feature query csv files
def read_feature_table(feature:str, csv_dir:Path) -> pd.DataFrame:
    filename = feature + ".csv.gz"
    filepath = csv_dir / filename

    try:
        feature_df = pd.read_csv(filepath)
    except FileNotFoundError:
        raise Exception(f"file {filename} dose not exist")

    feature_df = feature_df.sort_values(by=['subject_id'])
    if 'charttime' in feature_df.columns:
        feature_df['charttime'] = pd.to_datetime(feature_df['charttime'])
    return feature_df


# read the downloaded cohort query csv files
def read_cohort_data(cohort_name:str, csv_dir:Path):
    filename = cohort_name + '.csv.gz'
    filepath = csv_dir / filename

    try:
        cohort_df = pd.read_csv(filepath, parse_dates=['starttime', 'endtime'])
    except FileNotFoundError:
        raise Exception(f"cohort {cohort_name} dose not exist")
    
    cohort_df = cohort_df.sort_values(by=['subject_id'])
    return cohort_df


# read the downloaded outcome task query csv files and merge to specified cohort
def read_cohort_task(task_name:str, cohort_name:str, csv_dir:Path) -> pd.DataFrame:
    df_cohort = read_cohort_data(cohort_name, csv_dir)

    task_table = {
        "mortality" : ("hospital_expire_flag", "mortality.csv.gz"),
        "los_icu" : ("los_icu", "los_icu.csv.gz"),
        "dialysis" : ("dialysis_active", "los_icu.csv.gz"),
        "inv_vent" : ("inv_vent_flag", "inv_vent.csv.gz"),
        "vancomycin_dose" : ("dose_due", "vancomycin_dose.csv.gz"),
        "vancomycin_level" : ("vancomycin_level", "vancomycin_level.csv.gz"),
        "heparin_dose" : ("dose_due", "heparin_dose.csv.gz")
    }

    try:
        task_label = task_table[task_name][0]
        df_task = pd.read_csv(csv_dir / task_table[task_name][1])
    except KeyError:
        raise Exception(f"task {task_name} dose not exist")

    df_cohort = pd.merge(df_cohort, df_task, how='left', on=['subject_id'])
    if 'charttime' in df_cohort.columns:
        df_cohort['charttime'] = pd.to_datetime(df_cohort['charttime'])
    return task_label, df_cohort


def filter_cohort_task(df_cohort, task_label, args):
    """ Filter cohorts and task tables by los, start_time and stop_time """
    # Create base cohorts
    df_cohort_filtered = df_cohort[(df_cohort.los >= args.filter_los)]
    df_cohort_filtered = df_cohort_filtered.drop(columns=['los'])

    if args.task == 'dialysis':
        df_cohort_filtered = df_cohort_filtered[
            (df_cohort_filtered.dialysis_present == 0) | (
                (df_cohort_filtered.dialysis_present == 1) &
                (df_cohort_filtered.dialysis_start_time > 24)
            )]
        df_cohort_filtered.drop(
            ['dialysis_present', 'dialysis_start_time'], axis=1, inplace=True
        )
    df_task_filtered = df_cohort_filtered[task_label]
    df_cohort_filtered.drop([task_label], axis=1, inplace=True)

    # Filter data into window of time
    df_cohort_data = join_cohort_data(
        df_cohort_filtered, args.csv_dir, args.filter_start_time,
        args.filter_stop_time
    )
    return df_cohort_data, df_task_filtered


def export_cohort_data(df_inputs, df_outputs, args):
    """ Export cohort and task tables to allow for better reprodcucibility """
    str_inputs = args.csv_dir / f"""cohort_all_data_los{args.filter_los:0.1f}_
                    span_{args.filter_start_time:0.1f}-
                    {args.filter_stop_time:0.1f}
                    .csv""".replace('\n', '').replace(' ', '')
    str_outputs = args.csv_dir / f"""task_{args.task}_
                    los{args.filter_los:0.1f}_
                    span_{args.filter_start_time:0.1f}-
                    {args.filter_stop_time:0.1f}
                    .csv""".replace('\n', '').replace(' ', '')

    df_inputs.to_csv(str_inputs)
    df_outputs.to_csv(str_outputs)


from task import VancomycinDosingDataProcessor
import pandas as pd
from IPython.display import display
import os
from pathlib import Path
import dotenv

env_file = 'D:/UT_Third_Year/2022_Summer/Benchmark_Dataset/clinical-benchmarks/environment_config.env'
dotenv.load_dotenv(env_file, override=True)

print('start creating vanco object')
vanco = VancomycinDosingDataProcessor()
print('vanco object created')

time_step = 4
agg = 'last'
print('start creating vanco task_df')
vanco.create_task_df(time_step, agg)
print('done')


"""Utilities to transform data from one format/ontology/unit to another."""

import pandas as pd


def get_bmi(wt, ht):
    """Convert weight (kg) and height (cm) into BMI"""
    wt = wt[['Weight']].dropna().reset_index().groupby(wt.index.name
                                                      )['Weight'].median()
    ht = ht[['Height']].dropna().reset_index().groupby(ht.index.name
                                                      )['Height'].median()

    ht = ht / 100

    bmi = wt / (ht * ht)
    bmi = pd.DataFrame(bmi, columns=['bmi'])
    return bmi


def convert_icd_to_ccs(icd_code, version=10):
    pass



import argparse
import os
import sys
import logging
import pandas as pd
import logging
from pathlib import Path

from clinical_benchmarks.model import predict_task
from clinical_benchmarks.extract import (
    check_bq_components, run_queries, copy_from_bq_to_bucket, download_gcs_file
)

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
    datefmt='%m/%d/%Y %H:%M:%S'
)
logger = logging.getLogger(__name__)


class EnvDefault(argparse.Action):
    """Argument parser class which inherits from environment and has a default value."""
    def __init__(self, envvar, required=True, default=None, **kwargs):
        if not default and envvar:
            if envvar in os.environ:
                default = os.environ[envvar]
        if required and default:
            required = False
        super(EnvDefault,
              self).__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)


def dir_path(string):
    return Path(string)


def parse_arguments(arguments=None):
    """Parse input arguments for entry points."""
    parser = argparse.ArgumentParser(
        description="clinical-benchmarks command line interface"
    )
    parser.add_argument(
        '--csv_dir',
        required=True,
        action=EnvDefault,
        envvar='MODEL_DIR',
        default='csv/',
        type=dir_path,
        help='Directory with CSV tables'
    )

    subparsers = parser.add_subparsers(dest="actions", title="actions")
    subparsers.required = True

    # Downloading
    filter_download = subparsers.add_parser(
        "download", help=("Download options for cohort data")
    )
    filter_download.add_argument(
        '--project',
        action=EnvDefault,
        envvar='GCP_PROJECT',
        required=True,
        help='Google Cloud project'
    )
    filter_download.add_argument(
        '--dataset',
        action=EnvDefault,
        envvar='BQ_DATASET',
        required=True,
        help='BigQuery dataset for creation of tables'
    )
    filter_download.add_argument(
        '--bucket',
        action=EnvDefault,
        envvar='GCS_BUCKET',
        required=True,
        help='Google Cloud Storage bucket for downloading data'
    )
    filter_download.add_argument(
        '--mimic_derived',
        action='store_true',
        help='Downloads all the mimic_derived variable tables'
    )

    # Data filtering options
    filter_opt = subparsers.add_parser(
        "model", help=("Model filtering options for cohort data")
    )

    filter_opt.add_argument(
        '--task',
        choices=['mortality', 'los_icu', 'dialysis', 'inv_vent'],
        required=True
    )
    filter_opt.add_argument(
        '--los',
        type=float,
        default=1,
        dest='filter_los',
        help='LOS measured in fractional days'
    )
    filter_opt.add_argument(
        '--start_time',
        type=float,
        default=0,
        dest='filter_start_time',
        help='Start time units are fractional days'
    )
    filter_opt.add_argument(
        '--stop_time',
        type=float,
        default=1,
        dest='filter_stop_time',
        help='Stop time units are fractional days'
    )

    filter_opt.add_argument(
        '--model_type', default='lr', choices=['lr', 'xgb'], dest='model_type'
    )

    return parser.parse_args(arguments)


def download(args):
    """Run queries on BigQuery to generate datasets, copy data to GCS, then download locally."""
    check_bq_components(args.project, args.dataset, args.bucket)
    run_queries(args.project, args.dataset)

    download_file_list = [
        'cohort.csv.gz', 'vancomycin_cohort.csv.gz', 'chemistry.csv.gz',
        'coagulation.csv.gz', 'complete_blood_count.csv.gz',
        'weight.csv.gz','height.csv.gz', 'vitalsign.csv.gz',
        'vancomycin_dose.csv.gz', 'vancomycin_level.csv.gz',
        'labevents.csv.gz', 'mortality.csv.gz', 'los_icu.csv.gz',
        'inv_vent.csv.gz', 'dialysis.csv.gz', 'heparin_cohort.csv.gz',
        'heparin_dose.csv.gz'
    ]
    for f in download_file_list:
        copy_from_bq_to_bucket(f, args.project, args.dataset, args.bucket)
        download_gcs_file(f, args.csv_dir, args.project, args.bucket)

    if args.mimic_derived:
        full_variable_list = [
            'age', 'antibiotic', 'bg', 'blood_differential', 'cardiac_marker',
            'charlson', 'chemistry', 'coagulation', 'complete_blood_count',
            'crrt', 'dobutamine', 'enzyme', 'epinephrine', 'gcs', 'height',
            'icp', 'inflammation', 'norepinephrine', 'oxygen_delivery',
            'phenylephrine', 'rhythm', 'rrt', 'suspicion_of_infection',
            'urine_output', 'vasopressin', 'ventilation', 'ventilator_setting',
            'vitalsign', 'weight_durations'
        ]
        var_dir = args.csv_dir / 'variables'
        for var in full_variable_list:
            f = f'{var}.csv.gz'
            copy_from_bq_to_bucket(f, args.project, args.dataset, args.bucket)
            download_gcs_file(f, var_dir, args.project, args.bucket)


def main(argv=sys.argv):
    """Entry point for package."""

    args = parse_arguments(argv[1:])

    if args.actions == 'download':
        download(args)
    elif args.actions == 'model':
        if args.filter_start_time > args.filter_stop_time:
            logger.warn("Filter start_time is later than filter stop_time")
        model, y_pred = predict_task(args)
    else:
        logger.warn('Unrecongnized command')


if __name__ == '__main__':
    main()


"""Measures for evaluating the performance of models."""

from sklearn.metrics import roc_auc_score
from sklearn.metrics import mean_squared_error
from math import sqrt

def model_accuracy(outcome, y_test, y_pred):
    if outcome == 'los_icu':
        model_score = sqrt(mean_squared_error(y_test, y_pred))
    elif outcome == 'mortality':
        model_score = roc_auc_score(y_test, y_pred)
    else:
        model_score = roc_auc_score(y_test, y_pred)

    return model_score


# Functions to pull in tables, filter timing and join to cohort
from datetime import date
from xmlrpc.client import DateTime
import pandas as pd
import logging
from pathlib import Path

logger = logging.getLogger(__name__)


def join_cohort_data(df_cohort, csv_dir, filter_start_time, filter_stop_time):
    df_cohort_index = df_cohort[['subject_id', 'hadm_id', 'stay_id', 'intime']]

    # use same variables for now for all outcomes, can select custom later
    dynamic_variable_list = [
        'bg', 'blood_differential', 'coagulation', 'chemistry',
        'complete_blood_count', 'enzyme', 'gcs', 'vitalsign', 'weight_durations'
    ]
    static_variable_list = ['age', 'charlson', 'height']
    df_cohort_all_data = df_cohort_index
    for var in static_variable_list:
        logger.info('Joining table: ' + var)
        df_var = read_variable(df_cohort_index, var, csv_dir)
        join_columns = ['subject_id', 'hadm_id', 'stay_id', 'intime']
        df_cohort_all_data = pd.merge(
            df_cohort_all_data,
            df_var,
            how='left',
            on=join_columns,
            suffixes=('', '_' + var)
        )

    # Loops through and combines all the dynamic data, filtering out time windows and averaging
    for var in dynamic_variable_list:
        logger.info('Joining table: ' + var)
        df_var = read_variable(df_cohort_index, var, csv_dir)
        df_var_filtered = filter_variable(
            df_var, var, filter_start_time, filter_stop_time
        )

        num_col = df_var_filtered.select_dtypes(['int64', 'float64','datetime64[ns]'])\
                                                .drop(join_columns, axis=1).columns
        # str_col = df_var_filtered.select_dtypes(['object']).columns

        agg_dict = {}
        num_lambda = ['min', 'max']
        # str_lambda = lambda x: ','.join(x.unique())    # string agg if we wanted to do it
        # for col in str_col:
        #     agg_dict.update({col:str_lambda})

        for col in num_col:
            agg_dict.update({col: num_lambda})
        join_columns = ['subject_id', 'hadm_id', 'stay_id', 'intime']
        df_var_agg = df_var_filtered.groupby(join_columns,
                                             as_index=False).agg(agg_dict)
        df_var_agg.columns = df_var_agg.columns.map(
            lambda x: x[0] if x[0] in join_columns else '_'.join(x)
        )  # join multiindex into single index
        df_cohort_all_data = pd.merge(
            df_cohort_all_data,
            df_var_agg,
            how='left',
            on=join_columns,
            suffixes=('', '_' + var)
        )

    return df_cohort_all_data


def update_column_name(column_name, suffix):
    if column_name in ['subject_id', 'hadm_id', 'stay_id', 'intime']:
        new_column_name = column_name
    else:
        new_column_name = column_name + suffix
    return new_column_name


def filter_variable(df, variable_name, filter_start_time, filter_stop_time):
    date_columns = get_date_columns(variable_name)
    # Add duration column if there is a start_time and end_time
    if date_columns[0] == 'starttime':
        df['duration_' +
           variable_name] = df[date_columns[1]] - df[date_columns[0]]

    # Check if var is a lab, give extra window 6 hours before and 1 day after
    if variable_name in [
        'complete_blood_count', 'chemistry', 'blood_differential',
        'coagulation', 'enzyme'
    ]:
        filter_start_time = filter_start_time - 1 / 4  # 6 hours earlier
        filter_stop_time = filter_stop_time + 1  # 1 day later

    # Check if timestamp within filter bounds
    df["time_diff"] = df[date_columns[0]] - df.intime
    df_filtered = df[(df.time_diff.dt.days >= filter_start_time) &
                     (df.time_diff.dt.days <= filter_stop_time)]

    # drop time_diff and date_column since no longer needed
    if date_columns[
        0] == 'admittime':  # keep this, only age uses this date_column

        df_filtered = df_filtered.drop(['time_diff'], axis=1)
    else:
        df_filtered = df_filtered.drop(['time_diff'] + date_columns, axis=1)
    return df_filtered


def get_date_columns(variable_name):
    if variable_name == 'age':
        date_columns = ['admittime']
    elif variable_name == 'charlson':
        date_columns = []
    elif variable_name == 'suspicion_of_infection':
        date_columns = [
            'antibiotic_time', 'susected_infection_time', 'culture_time'
        ]
    elif variable_name in ['antibiotic', 'heparin']:
        date_columns = ['starttime', 'stoptime']
    elif variable_name in (
        [
            'dobutamine', 'epinephrine', 'norepinephrine', 'phenylephrine',
            'vasopressin', 'ventilation', 'weight_durations'
        ]
    ):

        date_columns = ['starttime', 'endtime']
    else:
        date_columns = ['charttime']
    return date_columns


def get_join_column(variable_name):
    if variable_name == 'age':
        join_columns = ['hadm_id', 'subject_id']
    elif variable_name in ['antibiotic', 'suspicion_of_infection']:
        join_columns = join_columns = ['stay_id', 'hadm_id', 'subject_id']
    elif variable_name in ['heparin', 'rhythm']:
        join_columns = ['subject_id']
    elif variable_name in [
        'crrt', 'dobutamine', 'epinephrine', 'norepinephrine', 'phenylephrine',
        'rrt', 'urine_output', 'vasopressin', 'ventilation', 'weight_durations',
        'suspicion_of_infection'
    ]:
        join_columns = ['stay_id']
    elif variable_name in [
        'gcs', 'height', 'icp', 'oxygen_delivery', 'ventilator_setting',
        'vitalsign'
    ]:
        join_columns = ['stay_id', 'subject_id']
    else:
        join_columns = ['hadm_id', 'subject_id']
    return join_columns


def read_variable(df_cohort, variable_name, csv_dir):
    date_columns = get_date_columns(variable_name)
    join_columns = get_join_column(variable_name)

    df_var = pd.read_csv(
        csv_dir / f"variables/{variable_name}.csv.gz", parse_dates=date_columns
    )

    df_var_with_time = pd.merge(df_cohort, df_var, how='left', on=join_columns)
    return df_var_with_time


def get_lab_features_by_stayid(source_df, stay_id, datetime, columns=None, window=24, agg='last'):
    if type(datetime) != 'datetime64[ns]':
        datetime = pd.to_datetime(datetime) # convert datetime format
    
    window_init_time = datetime - pd.Timedelta(hours=window) # obtain a window

    idx = source_df['stay_id'] == stay_id
    features = source_df.loc[idx]

    # get the features between the time window
    #idx = features['charttime'] >= window_init_time and source_df['charttime'] <= datetime
    idx = features['charttime'].between(window_init_time, datetime)
    features = features.loc[idx]

    if agg is not None:
        features = features.groupby('label').agg(agg)

    if columns is not None:
        features = features.loc[:, columns]
    
    return features


import pandas as pd
import os
import numpy as np
from pathlib import Path
from utils import read_cohort_task, read_feature_table, read_cohort_data

# rename to preprocessor
# drop the task name, output the final dataframe to a file
class BaseDataProcessor(object):
    def __init__(self):
        # load environment variables and necessary query tables
        self.csv_dir = Path(os.environ['MODEL_DIR']) # read dir path from environment variable
        self.chemistry_feature = read_feature_table('chemistry', self.csv_dir)
        self.coagulation_feature = read_feature_table('coagulation', self.csv_dir)
        self.blood_feature = read_feature_table('complete_blood_count', self.csv_dir)
        self.weight_feature = read_feature_table('weight', self.csv_dir)
        self.height_feature = read_feature_table('height', self.csv_dir)
        self.vitalsign_feature = read_feature_table('vitalsign', self.csv_dir)
        self.vitalsign_feature = self.vitalsign_feature.drop_duplicates().reset_index(drop=True)

    '''
        @param: {start, stop} two integer represent how long a time window is
        @param: {agg} a method of aggregation, ie. last
        @param: {feature} feature that need extract and aggregate
        @param: {columns} specific columns that want return
        @return aggregated features in a specified time offset between start and stop
                the time offset is counted from the ICU intime
    '''
    def get_feature_data(self, start:int, stop:int, agg:str, feature:pd.DataFrame, columns=None) -> pd.DataFrame:
        start = pd.Timedelta(start, 'h')
        stop = pd.Timedelta(stop, 'h') # convert to hour for computing time offset

        # convert chartime to event time, depend on event type
        feature['charttime'] = pd.to_datetime(feature['charttime'])
        intime_col = self.task_df[['subject_id', 'intime', 'outtime']]
        intime_col = intime_col.drop_duplicates()

        feature_df = feature.merge(intime_col, how='inner', on=['subject_id']) # add in the intime col to compute offset

        # filter out the lab result dose belong to current ICU admission/duration
        feature_df['valid'] = feature_df.apply(lambda x: 1 if (x['charttime'] > x['intime'] and x['charttime'] < x['outtime']) else 0, axis=1)
        idx = feature_df['valid'] != 1
        feature_df = feature_df.drop(feature_df.loc[idx].index, axis=0) # drop the records outside ICU admission
        feature_df = feature_df.drop('valid', axis=1)

        # compute time offset column for each feature since ICU admission time / intime
        feature_df['event_time_offset'] = feature_df['charttime'] - feature_df['intime']
        feature_df = feature_df.drop(['intime', 'outtime'], axis=1)

        # locate the feature event that happens during desired period
        idx = feature_df['event_time_offset'].between(start, stop)
        feature_df = feature_df.loc[idx].sort_values(by='charttime')

        # aggregate the features for each subject
        feature_df = feature_df.groupby(['subject_id']).agg(agg)

        if columns is not None:
            feature_df = feature_df.loc[:, columns]
        
        return feature_df


    '''
        @param: {cohort_df} a subject cohort from a task object
        @param: {timestep} a str value stated in hour, represent the increment of time from ICU starttime to endtime
        @return: a time series dataframe with timestep increment for each subject
    '''
    def create_time_cohort(self, cohort_df:pd.DataFrame, time_step:int) -> pd.DataFrame:
        time_step = str(time_step) + 'H'
        cohort_with_time = pd.DataFrame({})
        cohort_df['starttime'] = pd.to_datetime(cohort_df['starttime'])
        cohort_df['endtime'] = pd.to_datetime(cohort_df['endtime'])
        for _, row in cohort_df.iterrows():
            to_insert = pd.date_range(row['starttime'], row['endtime'], freq=time_step)
            to_insert = pd.DataFrame(to_insert)
            to_insert['subject_id'] = row['subject_id']
            to_insert['episode_id'] = row['episode_id']
            cohort_with_time = pd.concat([cohort_with_time, to_insert])

        cohort_with_time = cohort_with_time.rename(columns={0 : 'time'})
        return cohort_with_time


    '''
        merge feature table into a time series cohort
        @param: {cohort_with_time} the outcome dataframe return by method {create_time_cohort}
        @param: {feature} the feature dataframe that want to merge in
        @param: {cohort_df} the cohort dataframe of the current Task object, ie. VancomycinDosing.cohort_df
        @prarm: {time_step} a integer that represent how many hours will last for a time window
        @return return a cohort_with_time dataframe that each time step matches with a feature measurement, might have None for a time step
    '''
    def merge_feature_to_time_dataframe(self, cohort_with_time:pd.DataFrame, feature:pd.DataFrame, cohort_df:pd.DataFrame, agg:str, time_step:int):
        feature_df = cohort_df.merge(feature, how='inner', on='subject_id')
        # first filter the feature events that happens outside the ICU stay time
        feature_df = feature_df.loc[(feature_df['charttime'] >= feature_df['starttime']) & (feature_df['charttime'] <= feature_df['endtime'])]
        feature_df = feature_df.sort_values(by='charttime').reset_index()
        feature_df = feature_df.drop(['starttime', 'endtime', 'index'], axis=1)

        res = cohort_with_time.merge(feature_df, how='inner', on=['subject_id', 'episode_id'])
        # match each feature measurement to the corresponding time window defined by the time_step
        # ie. between 4:00 to 4:00 + time_step
        res = res.loc[(res['charttime'] >= res['time']) & (res['charttime'] <= res['time'] + pd.Timedelta(time_step, 'h'))]
        res = res.groupby(by=['episode_id', 'time']).agg(agg)
        res = cohort_with_time.merge(res, how='left', on=['subject_id', 'time'])

        res = res.drop('charttime', axis=1)
        return res
    

    """
        functionality similar to {merge_feature_to_time_dataframe}, but weight features are recorded based on duration
        need different ways of processing and merging
    """
    def merge_weight_to_time_dataframe(self, cohort_with_time:pd.DataFrame, weight_feature:pd.DataFrame, cohort_df:pd.DataFrame, agg:str):
        cohort = cohort_df[['subject_id', 'episode_id']]
        # filter out the weight feature that not belongs to current cohort
        feature_df = cohort.merge(weight_feature, how='inner', on='subject_id')

        res = cohort_with_time.merge(feature_df, how='inner', on=['subject_id', 'episode_id'])
        # match each time step to weight durations
        res = res.loc[res['time'].between(res['starttime'], res['endtime'])]
        res = res.groupby(by=['episode_id', 'time']).agg(agg)
        res = cohort_with_time.merge(res, how='left', on=['subject_id', 'time'])
        res = res.drop(['starttime', 'endtime'], axis = 1)
        return res


    """
        functionality similar to {merge_feature_to_time_dataframe}, but height features are less frequently recorded
        need different ways of processing and merging
    """
    def merge_height_to_time_dataframe(self, cohort_with_time:pd.DataFrame, height_feature:pd.DataFrame, cohort_df:pd.DataFrame, agg:str):
        feature_df = cohort_df.merge(height_feature, how='inner', on='subject_id')
        # filter out the height features that measured outside ICU stay
        # but lots of subjects only have one height measurement, usually 3-5 minutes before ICU starttime
        # extend the window by 5 minutes to include more height features
        feature_df = feature_df.loc[(feature_df['charttime'] >= (feature_df['starttime'] - pd.Timedelta(5, 'm'))) & (feature_df['charttime'] <= feature_df['endtime'])]
        feature_df = feature_df.drop(['starttime', 'endtime'], axis=1)

        res = cohort_with_time.merge(feature_df, how='left', on=['subject_id', 'episode_id'])
        res = res.drop('charttime', axis=1)
        return res
    

    """
        functionality similar to {merge_feature_to_time_dataframe}, but the size of vitalsign dataframe is huge
        It has to be sliced into several chunks and process separately, otherwise may exceed memory limit
    """
    def merge_vitalsign_to_time_dataframe(self, cohort_with_time:pd.DataFrame, vitalsign_feature:pd.DataFrame, cohort_df:pd.DataFrame, agg:str, time_step:int):
        feature_df = cohort_df.merge(vitalsign_feature, how='inner', on='subject_id')
        # first filter the feature events that happens outside the ICU stay time
        feature_df = feature_df.loc[(feature_df['charttime'] >= feature_df['starttime']) & (feature_df['charttime'] <= feature_df['endtime'])]
        feature_df = feature_df.sort_values(by='charttime').reset_index()
        feature_df = feature_df.drop(['starttime', 'endtime', 'index'], axis=1)

        # determine the number of rows, and horizontally slice the dataframe into several chunks
        # process each chunk independently to avoid exceeding memory limit
        process_list = []
        length = feature_df.shape[0]
        num_chunks = 30
        interval = int(length * (1/num_chunks))

        # slice the dataframe into chunks
        cur = 0
        for i in range(1, num_chunks):
            process_list.append(feature_df.loc[cur : cur + interval])
            cur += interval
        process_list.append(feature_df.loc[cur : length])

        for i in range(len(process_list)):
            res = cohort_with_time.merge(process_list[i], how='inner', on=['subject_id', 'episode_id'])
            # match each event to its closest time slot
            res = res.loc[(res['charttime'] >= res['time']) & (res['charttime'] <= res['time'] + pd.Timedelta(time_step, 'h'))]
            res = res.groupby(by=['episode_id', 'time']).agg(agg)
            process_list[i] = res
        
        # concat the chuncks back into a single dataframe
        res = pd.concat(process_list)
        res = cohort_with_time.merge(res, how='left', on=['subject_id', 'time'])
        # regroup to remove multiple matches
        res = res.groupby(by=['time', 'subject_id', 'episode_id']).agg(agg)
        res = res.sort_values(['subject_id', 'episode_id', 'time']).reset_index().drop('charttime', axis=1)
        return res


    '''
        functionality similar to {merge_feature_to_time_dataframe}, but used to merge action/reward into time series cohort
        @param: {action_reward} a task_df for reinforcement learning task. ie. VancomycinDosing.task_df
    '''
    def merge_action_to_time_dataframe(self, cohort_with_time:pd.DataFrame, action_reward:pd.DataFrame, agg:str, time_step:int):
        res = cohort_with_time.merge(action_reward, how='inner', on=['subject_id', 'episode_id'])
        # match each action/reward to its closest time step by charttime
        res = res.loc[(res['charttime'] >= res['time'] - pd.Timedelta(time_step, 'h')) & (res['charttime'] <= res['time'] )]
        # if multiple match, only take last one
        res = res.groupby(by=['episode_id', 'time']).agg(agg)
        res = cohort_with_time.merge(res, how='left', on=['subject_id', 'episode_id', 'time'])
        res = res.drop('charttime', axis=1)
        return res



# a Vancomycin Dosing Prediction Task object
# have a create method to create df
# have a save method to save final df
class VancomycinDosingDataProcessor(BaseDataProcessor):
    def __init__(self):
        super(VancomycinDosingDataProcessor, self).__init__()
        self.cohort_name = "vancomycin_cohort"
        vancomycin_cohort = read_cohort_data(self.cohort_name, self.csv_dir)
        task_label, vancomycin_dose = read_cohort_task('vancomycin_dose', self.cohort_name, self.csv_dir)
        task_label, vancomycin_level = read_cohort_task('vancomycin_level',self.cohort_name, self.csv_dir)

        vancomycin_dose['dose_due'] = vancomycin_dose.apply(lambda x : 0 if x['event_txt'] == 'Hold Dose' else x['dose_due'], axis=1)
        vancomycin_cohort = vancomycin_cohort.drop('los', axis=1)
        self.cohort_df = vancomycin_cohort.sort_values(by=['subject_id'], axis=0).reset_index().drop('index', axis=1)
        self.vancomycin_dose = vancomycin_dose
        self.vancomycin_level = vancomycin_level


    '''
        create the task dataframe for vancomycin dosing reinforcement learning task
    '''
    def create_task_df(self, time_step:int, agg:str)->pd.DataFrame:
        vancomycin_dose = self.vancomycin_dose
        vancomycin_level = self.vancomycin_level

        vancomycin_dose = vancomycin_dose.drop(['dose_given', 'dose_given_unit', 'medication'], axis=1)

        vancomycin_dose.rename(columns={'dose_due':'action'}, inplace=True)
        vancomycin_level.rename(columns={'vancomycin_level':'reward'}, inplace=True)

        # concat dose and measure tables together, then sort based on charttime
        task_df = pd.concat([vancomycin_dose, vancomycin_level])
        task_df = task_df.sort_values(by=['subject_id', 'charttime', 'dose_num'], axis=0)
        task_df = task_df.reset_index().drop('index', axis=1)
        self.task_df = task_df[['subject_id', 'episode_id', 'charttime', 'dose_num', 'action', 'reward']]

        # use the glucose from the vitalsign table instead of chemistry
        self.chemistry_feature = self.chemistry_feature.drop('glucose', axis=1)
        # merge all features together with task_df
        self.cohort_with_time = self.create_time_cohort(self.cohort_df, time_step)

        feature_table = {
            'chemistry':self.chemistry_feature, 
            'coagulation':self.coagulation_feature, 
            'vitalsign':self.vitalsign_feature,
            'blood':self.blood_feature, 
            'height':self.height_feature, 
            'weight':self.weight_feature
            }
        
        action_reward_cohort = self.merge_action_to_time_dataframe(self.cohort_with_time, self.task_df, agg, time_step)

        for name, feature in feature_table.items():
            if name == 'height':
                temp = self.merge_height_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg)
            elif name == 'weight':
                temp = self.merge_weight_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg)
            elif name == 'vitalsign':
                temp = self.merge_vitalsign_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg, time_step)
            else:
                temp = self.merge_feature_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg, time_step)
            action_reward_cohort = action_reward_cohort.merge(temp, how='left', on=['subject_id', 'episode_id', 'time'])
            del temp
        
        self.featured_cohort_with_time = action_reward_cohort
    

    '''
        automatically save the task dataframe to the given directory in .csv.gz format for later reuse
    '''
    def save_task_df(self, csv_dir:Path, filename:str):
        if 'csv.gz' not in filename:
            raise Exception(f"desired filename with extension .csv.gz, not {filename}")
        
        save_path = csv_dir / filename
        self.featured_cohort_with_time.to_csv(save_path, index=False, compression="gzip")


class HeparinDosingDataProcessor(BaseDataProcessor):
    def __init__(self):
        super(HeparinDosingDataProcessor, self).__init__()
        self.cohort_name = 'heparin_cohort'
        self.cohort_df = read_cohort_data(self.cohort_name, self.csv_dir).drop('los', axis=1)

        task_label, heparin_dose = read_cohort_task('heparin_dose', self.cohort_name, self.csv_dir)
        heparin_dose = heparin_dose.drop(['dose_given', 'dose_given_unit', 'medication', 'starttime', 'endtime', 'los'], axis=1)
        # set the dose amount to 0 for the 'Hold Dose' case
        heparin_dose['dose_due'] = heparin_dose.apply(lambda x : 0 if x['event_txt'] == 'Hold Dose' else x['dose_due'], axis=1)
        self.heparin_dose = heparin_dose.sort_values(by=['subject_id', 'dose_num'], axis=0).reset_index(drop=True)

        heparin_level = self.coagulation_feature.drop(['pt', 'ptt'], axis=1)
        heparin_level = self.cohort_df.merge(heparin_level, how='left', on=['subject_id'])
        # filter out the measurements that are outside of current ICU stay
        heparin_level = heparin_level.loc[(heparin_level['charttime'] > heparin_level['starttime']) & (heparin_level['charttime'] < heparin_level['endtime'])]
        self.heparin_level = heparin_level.drop(['starttime', 'endtime'], axis=1)
    

    def create_task_df(self, time_step:int, agg:str):
        heparin_dose = self.heparin_dose
        heparin_level = self.heparin_level

        heparin_dose.rename(columns={'dose_due':'action'}, inplace=True)
        heparin_level.rename(columns={'inr':'reward'}, inplace=True)

        # concat dose and measure tables together, then sort based on charttime
        task_df = pd.concat([heparin_dose, heparin_level])
        task_df = task_df.sort_values(by=['subject_id', 'charttime', 'dose_num'], axis=0)
        task_df = task_df.reset_index(drop=True)

        # sperate Bolus dose and Infusion dose into two different actions
        task_df['administration_type'] = task_df['administration_type'].apply(lambda x : None if isinstance(x, type(np.nan)) else x)
        task_df['administration_type'] = task_df['administration_type'].apply(lambda x : 'None' if isinstance(x, type(None)) else x)

        task_df['action_bolus'] = task_df.loc[task_df['administration_type'].str.contains('Bolus|Maintenance')]['action']
        task_df['action_infusion'] = task_df.loc[task_df['administration_type'].str.contains('Infusion')]['action']
        self.task_df = task_df.drop(['administration_type', 'action', 'dose_due_unit', 'event_txt'], axis=1)

        # use the glucose from the vitalsign table instead of chemistry
        self.chemistry_feature = self.chemistry_feature.drop('glucose', axis=1)
        # merge all features together with task_df
        self.cohort_with_time = self.create_time_cohort(self.cohort_df, time_step)

        feature_table = {
            'chemistry':self.chemistry_feature, 
            'vitalsign':self.vitalsign_feature,
            'blood':self.blood_feature, 
            'height':self.height_feature, 
            'weight':self.weight_feature
            }

        action_reward_cohort = self.merge_action_to_time_dataframe(self.cohort_with_time, self.task_df, agg, time_step)

        for name, feature in feature_table.items():
            if name == 'height':
                temp = self.merge_height_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg)
            elif name == 'weight':
                temp = self.merge_weight_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg)
            elif name == 'vitalsign':
                temp = self.merge_vitalsign_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg, time_step)
            else:
                temp = self.merge_feature_to_time_dataframe(self.cohort_with_time, feature, self.cohort_df, agg, time_step)
            action_reward_cohort = action_reward_cohort.merge(temp, how='left', on=['subject_id', 'episode_id', 'time'])
            del temp
        
        self.featured_cohort_with_time = action_reward_cohort


    def save_task_df(self, csv_dir:Path, filename:str):
        if 'csv.gz' not in filename:
            raise Exception(f"desired filename with extension .csv.gz, not {filename}")
        
        save_path = csv_dir / filename
        self.featured_cohort_with_time.to_csv(save_path, index=False, compression="gzip")


"""Classes and utilities for training machine learning models."""
import logging
from sklearn.pipeline import Pipeline
from sklearn.model_selection import train_test_split
from sklearn.linear_model import (LogisticRegression, LinearRegression)
import xgboost as xgb

from clinical_benchmarks.evaluate import model_accuracy
from clinical_benchmarks.preprocess import custom_preprocessor
from clinical_benchmarks.utils import read_cohort_task
from clinical_benchmarks.utils import filter_cohort_task
from clinical_benchmarks.utils import export_cohort_data

logger = logging.getLogger(__name__)


def train_task(X_train, y_train, preprocessor, task, model_type):
    if (model_type == 'lr') & (task == 'los_icu'):
        mdl = LinearRegression()
    elif model_type == 'lr':
        mdl = LogisticRegression(max_iter=1000)
    elif model_type == 'xgb':
        mdl = xgb.XGBRegressor()
        params = {
            'learning_rate': 0.1,
            'gamma': 0.5,
            'max_depth': 10,
            'colsample_bytree': 0.8,
            'subsample': 0.9,
            'min_child_weight': 5,
            'n_estimators': 250,
            'eval_metric': 'logloss',
            'use_label_encoder': False
        }
        mdl.set_params(**params)
    else:
        logger.error(f"Model type '{model_type}' is invalid")

    pipeline = Pipeline(steps=[('preprocessor', preprocessor), ('model', mdl)])
    model = pipeline.fit(X_train, y_train.values.ravel())
    return pipeline, model


def predict_task(args):
    task_label, df_cohort = read_cohort_task(args.task, args.cohort_name, args.csv_dir)
    df_inputs, df_outputs = filter_cohort_task(df_cohort, task_label, args)

    preprocessor = custom_preprocessor(df_inputs, df_outputs, args.model_type)
    X_train, X_test, y_train, y_test = train_test_split(
        df_inputs, df_outputs, test_size=0.25, random_state=42
    )
    pipeline, model = train_task(
        X_train, y_train, preprocessor, args.task, args.model_type
    )
    y_pred = pipeline.predict(X_test)

    model_acc = model_accuracy(args.task, y_test, y_pred)
    logger.info(f'{args.task} accuracy: {model_acc:0.3f}')
    # write out cohort data used
    export_cohort_data(df_inputs, df_outputs, args)
    return model, y_pred, y_test



"""Convenience classes for loading and working with datasets."""
import torch
import pandas as pd


class SimpleDataset(torch.utils.data.Dataset):
    def __init__(self, X: pd.DataFrame, y: pd.DataFrame):
        self.X = X
        self.y = y

    def __getitem__(self, idx):
        return torch.tensor(self.X.loc[idx]), self.y.loc[idx]

    def __len__(self):
        return self.X.shape[0]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='clinical_benchmarks',
    version='1.0.0',
    author="Alex Bennett, Pengcheng Wang",
    author_email="alexmbennett2@gmail.com, cheng666999@gmail.com",
    description="A repo of clinical benchmarks for MIMIC-IV",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/kind-lab/clinical-benchmarks",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts':
        ['clinical_benchmarks = clinical_benchmarks.__main__:main']
    },
)



"""Utilities for preprocessing extracted datasets."""
import numpy as np
import pandas as pd
from scipy import stats

from sklearn.pipeline import Pipeline
from sklearn.preprocessing import (
    StandardScaler, OneHotEncoder, OrdinalEncoder
)
from sklearn.impute import SimpleImputer
from sklearn.compose import ColumnTransformer
from sklearn.base import TransformerMixin


class DropMissingFeatures(TransformerMixin):
    def __init__(self, thresh):
        self.thresh = thresh
        self.drop_features = []

    def fit(self, X, y):
        self.drop_features = [
            c for c in X if X[c].isnull().sum() / len(X) > self.thresh
        ]
        return self

    def transform(self, X, y=None):
        X_ = X.drop(columns=self.drop_features)
        return X_


class DropSpecificFeatures(TransformerMixin):
    def __init__(self, drop_features):
        self.drop_features = drop_features

    def fit(self, X, y):
        # check that features are actually in X
        self.drop_features = [c for c in self.drop_features if c in X]
        return self

    def transform(self, X, y=None):
        X_ = X.drop(columns=self.drop_features)
        return X_


def custom_preprocessor(df_inputs, df_outputs, model_type):
    categorical_features = df_inputs.select_dtypes(include=['object']).columns
    numeric_features = df_inputs.select_dtypes(
        include=['int64', 'float64']
    ).columns

    transformers = []
    numeric_transformer = Pipeline(
        steps=[
            ('drop_missing', DropMissingFeatures(0.98)),
            ('drop_specific', DropSpecificFeatures(['fio2_min', 'fio2_max'])),
            ('imputer', SimpleImputer(missing_values=np.nan, strategy='mean')
            ), ('scaler', StandardScaler())
        ]
    )
    transformers.append(('num', numeric_transformer, numeric_features))

    if model_type == 'xgb':
        encoder = OrdinalEncoder()
    else:
        encoder = OneHotEncoder(handle_unknown='ignore')
    categorical_transformer = Pipeline(
        steps=[
            ('drop_missing', DropMissingFeatures(0.98)),
            (
                'imputer',
                SimpleImputer(strategy='constant', fill_value='MISSING')
            ), ('encoder', encoder)
        ]
    )
    transformers.append(('cat', categorical_transformer, categorical_features))

    return ColumnTransformer(transformers=transformers)


def drop_feature_na_percent(df, na_percent=0.98, drop_features=[]):
    """ Remove features with greater than na_percent missing data """
    drop_columns = [
        c for c in df if df[c].isnull().sum() / len(df) > na_percent
    ]
    drop_columns.extend(drop_features)
    df = df.drop(columns=drop_columns)
    return df


def impute_using_additive_formula(df, a, b, c, verbose=True):
    """
    If we know a = b + c, we can impute data
    Many features are related in this way
    e.g. Total protein, albumin, globulin
    
    This is useful for merging features and reducing redundancy in the
    feature space.
    """

    if verbose:
        print(f'{a} = {b} + {c}')
    # === impute a ===
    # first we need b and c to not be null
    idx = df[b].notnull() & df[c].notnull()
    # rows where a is null
    idx = idx & df[a].isnull()
    # a = b + c
    df.loc[idx, a] = df.loc[idx, b] + df.loc[idx, c]
    if verbose:
        n = idx.sum()
        print(f'{a} - Imputed {n} values')

    # === impute b ===
    idx = df[a].notnull() & df[c].notnull()
    # rows where b is null
    idx = idx & df[b].isnull()
    # b = a - c
    df.loc[idx, b] = df.loc[idx, a] - df.loc[idx, c]
    if verbose:
        n = idx.sum()
        print(f'{b} - Imputed {n} values')

    # === impute c ===
    idx = df[a].notnull() & df[b].notnull()
    # rows where c is null
    idx = idx & df[c].isnull()
    # c = a - b
    df.loc[idx, c] = df.loc[idx, a] - df.loc[idx, b]
    if verbose:
        n = idx.sum()
        print(f'{c} - Imputed {n} values')

    return df


def _q25(x):
    """
    Compute percentile (25th)
    """
    return np.nanpercentile(x.values, 25)


def _q75(x):
    """
    Compute percentile (75th)
    """
    return np.nanpercentile(x.values, 75)


def _std(x, ddof=1):
    """
    Compute standard deviation with ddof degrees of freedom
    """
    return np.nanstd(x.values, ddof=ddof)


def _range(x):
    """
    Compute range (max-min)
    """
    return max(x) - min(x)


def _first(x):
    """
    Get first
    """
    return x.iloc[0]


def _last(x):
    """
    Get last
    """
    return x.iloc[-1]


def _slope(df, timecol):
    """
    get the slope
    """
    d = {}
    for col in df.columns:
        if col == timecol:
            continue
        ts = df[[timecol, col]].dropna()
        x = ts[timecol]
        y = ts[col]
        try:
            slope, intercept, r_val, p_val, std_err = stats.linregress(x, y)
        except ValueError:
            slope = np.nan
        d[f'_slope_{col}'] = slope

    return pd.Series(d)
